<?php
/**
 * @file
 * Contains \Drupal\payment_authnet\Form\TestForm.
 */

namespace Drupal\payment_authnet\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\payment_authnet\Entity\AuthnetProfile;
use Drupal\payment_authnet\Exception;
use net\authorize\api\contract\v1\AuthenticateTestRequest;
use net\authorize\api\controller\AuthenticateTestController;

/**
 * Defines a form to configure maintenance settings for this site.
 */
class TestForm extends FormBase {

  /**
   * The list of Authnet Profiles.
   *
   * @var array
   */
  protected $authnetProfiles;

  /**
   * Constructs a new TestForm object.
   */
  public function __construct() {
    $this->authnetProfiles = AuthnetProfile::loadMultiple();
  }

  /**
   * {@inheritdoc}
   */
  public function getFormID() {
    return 'payment_authnet_test';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $options = array('' => $this->t('- Select a profile -'));
    foreach ($this->authnetProfiles as $id => $authnet_profile) {
      $options[$id] = $authnet_profile->label();
    }
    $profile_ids = array_keys($this->authnetProfiles);

    $form['profileId'] = array(
      '#title' => $this->t('Authorize.net profile'),
      '#type' => 'select',
      '#options' => $options,
      '#default_value' => count($profile_ids) ? reset($profile_ids) : '',
    );

    $form['actions'] = array(
      'submit' => array(
        '#type' => 'submit',
        '#value' => t('Test'),
      )
    );

    return $form;
  }

  /**
   * {@inheritdoc}
   *
   * @todo Proper exception handling. Plural messages.
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $profileId = $form_state->getValue('profileId');

    if (empty($profileId)) {
      drupal_set_message(t('You need to select a profile to test.'), 'error');
      return;
    }
    $profile = $this->authnetProfiles[$profileId];

    try {
      // Common setup for API credentials.
      $refId = $profile->getApiKey() . REQUEST_TIME;

      $testRequest = new AuthenticateTestRequest();
      $testRequest->setMerchantAuthentication($profile->getMerchantAuthentication());
      $testRequest->setRefId($refId);

      $controller = new AuthenticateTestController($testRequest);
      $response = $controller->executeWithApiResponse($profile->getApiEnvironment());

      if ($response == NULL) {
        throw new Exception\PaymentAuthnetApiException($this->t('Null response returned.'));
      }
      if ($response->getRefId() != $refId) {
        throw new Exception\PaymentAuthnetApiException($this->t('Invalid Reference ID in response.'));
      }
      if ($response->getMessages()->getResultCode() != 'Ok') {
        throw new Exception\PaymentAuthnetApiException($this->t('Response returned @resultCode. @message (@code)', [
          '@resultCode' => $response->getMessages()->getResultCode(),
          '@message' => $response->getMessages()->getMessage()[0]->getText(),
          '@code' => $response->getMessages()->getMessage()[0]->getCode(),
        ]));
      }

      drupal_set_message($this->t('Response returned @resultCode. @message (@code).', [
        '@resultCode' => $response->getMessages()->getResultCode(),
        '@message' => $response->getMessages()->getMessage()[0]->getText(),
        '@code' => $response->getMessages()->getMessage()[0]->getCode(),
      ]));
    }
    catch (Exception\PaymentAuthnetApiException $e) {
      drupal_set_message($e->getMessage(), 'error');
    }
    catch (Exception\PaymentAuthnetSdkException $e) {
      drupal_set_message($e->getMessage(), 'error');
    }
    catch (Exception $e) {
      \Drupal::logger('payment_authnet')->error($e->getMessage());
    }

    return;
  }
}
