<?php

/**
 * @file
 * Contains \Drupal\payment_authnet\Plugin\Payment\Method\Authnet.
 */

namespace Drupal\payment_authnet\Plugin\Payment\Method;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Utility\Token;
use Drupal\payment\Entity\PaymentInterface;
use Drupal\payment\EventDispatcherInterface;
use Drupal\payment\Plugin\Payment\Status\PaymentStatusManagerInterface;
use Drupal\payment\Plugin\Payment\Method\PaymentMethodBase;
use Drupal\payment\Plugin\Payment\Method\PaymentMethodCapturePaymentInterface;
use Drupal\payment\Plugin\Payment\Method\PaymentMethodRefundPaymentInterface;
use Drupal\payment\Plugin\Payment\Method\PaymentMethodUpdatePaymentStatusInterface;
use Drupal\payment_authnet\Entity\AuthnetProfile;
use Drupal\payment_authnet\Exception;
use net\authorize\api\contract\v1 as AnetAPI;
use net\authorize\api\controller as AnetController;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * A payment method using Authnet.
 *
 * Plugins that extend this class must have the following keys in their plugin
 * definitions:
 * - entity_id: (string) The ID of the payment method entity the plugin is for.
 * - execute_status_id: (string) The ID of the payment status plugin to set at
 *   payment execution.
 * - capture: (boolean) Whether or not payment capture is supported.
 * - capture_status_id: (string) The ID of the payment status plugin to set at
 *   payment capture.
 * - refund: (boolean) Whether or not payment refunds are supported.
 * - refund_status_id: (string) The ID of the payment status plugin to set at
 *   payment refund.
 *
 * @PaymentMethod(
 *   deriver = "Drupal\payment_authnet\Plugin\Payment\Method\AuthnetDeriver",
 *   id = "payment_authnet",
 *   operations_provider = "\Drupal\payment_authnet\Plugin\Payment\Method\AuthnetOperationsProvider",
 * )
 */
class Authnet extends PaymentMethodBase implements ContainerFactoryPluginInterface, PaymentMethodCapturePaymentInterface, PaymentMethodRefundPaymentInterface, PaymentMethodUpdatePaymentStatusInterface {

  /**
   * The payment status manager.
   *
   * @var \Drupal\payment\Plugin\Payment\Status\PaymentStatusManagerInterface
   */
  protected $paymentStatusManager;

  /**
   * The AnetAPI payment object (contains info about credit card).
   *
   * @var \AnetAPI\PaymentType.
   */
  protected $AnetApiPayment;

  /**
   * The AnetAPI transaction request object (contains info about transaction).
   *
   * @var \AnetAPI\TransactionRequestType.
   */
  protected $transactionRequest;

  /**
   * The AnetAPI transaction request Type.
   *
   * @var string.
   */
  protected $transactionType;

  /**
   * The AnetAPI transaction request Type.
   *
   * @var \AnetAPI\TransactionDetailsType.
   */
  protected $transactionDetails;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * Custom Fields Values
   *
   * @var array
   */
  protected $customFieldsValues;

  /**
   * Authnet Profile.
   *
   * @var \Drupal\payment_authnet\Entity\AuthnetProfileInterface
   */
  protected $AuthnetProfile;

  /**
   * Constructs a new instance.
   *
   * @param mixed[] $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed[] $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   * @param \Drupal\payment\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   * @param \Drupal\Core\Utility\Token $token
   *   The token API.
   * @param \Drupal\payment\Plugin\Payment\Status\PaymentStatusManagerInterface
   *   The payment status manager.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity manager.
   */
  public function __construct(array $configuration, $plugin_id, array $plugin_definition, ModuleHandlerInterface $module_handler, EventDispatcherInterface $event_dispatcher, Token $token, PaymentStatusManagerInterface $payment_status_manager, EntityFieldManagerInterface $entity_field_manager) {
    $configuration += $this->defaultConfiguration();
    parent::__construct($configuration, $plugin_id, $plugin_definition, $module_handler, $event_dispatcher, $token, $payment_status_manager);
    $this->entityFieldManager = $entity_field_manager;
    $this->AuthnetProfile = AuthnetProfile::load($this->pluginDefinition['profile']);
    $this->setAnetApiPayment();
    $this->transactionType = PAYMENT_AUTHNET_TRANSACTION_AUTHORIZE;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('module_handler'),
      $container->get('payment.event_dispatcher'),
      $container->get('token'),
      $container->get('plugin.manager.payment.status'),
      $container->get('entity_field.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $elements = parent::buildConfigurationForm($form, $form_state);

    // Add credit card fields.
    $elements['credit_card_number'] = array(
      '#type' => 'creditfield_cardnumber',
      '#title' => $this->t('Credit Card Number'),
      '#description' => $this->t('Your 16 digits credit card number.'),
      '#maxlength' => 16,
      '#size' => 16,
      '#required' => TRUE,
    );

    $elements['expiration_date'] = array(
      '#type' => 'creditfield_expiration',
      '#title' => $this->t('Expiration Date'),
      '#required' => TRUE,
    );

    $elements['credit_card_cvv'] = array(
      '#type' => 'creditfield_cardcode',
      '#title' => $this->t('CVV Code'),
      '#maxlength' => 4,
      '#size' => 4,
      '#description' => $this->t('Your 3 or 4 digit security code on the back of your card.'),
      '#required' => TRUE,
    );
    return $elements;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    // Create the payment data for a credit card.
    $card = $form_state->getValues()['container']['plugin_form'];
    $this->setCreditCard($card['credit_card_number'], $card['expiration_date'], $card['credit_card_cvv']);

    // Save last 4 digits of credit card and expiration date in database.
    // This is useful for refund operations.
    unset($card['credit_card_cvv']);
    $card['credit_card_number'] = substr($card['credit_card_number'], -4);
    $this->setConfiguration($card);

    // Temporarily save other form fields values.
    // @todo Find a better way to process form fields values.
    // @see https://www.drupal.org/project/payment/issues/2841035.
    $fs_values = $form_state->getValues();
//    $list = new \Drupal\Core\Field\FieldItemList($fs_values);
//    $list->getValue();
    foreach ($this->entityFieldManager->getFieldDefinitions('payment', $this->getPayment()->bundle()) as $field_name => $field_definition) {
      if (!$field_definition->getFieldStorageDefinition()->isBaseField() && isset($fs_values[$field_name])) {
        if (count($fs_values[$field_name]) > 1) {
          foreach ($fs_values[$field_name] as $key => $field) {
            $this->customFieldsValues[$field_name][$key] = reset($field);
          }
        }
        else {
          $first_value = reset($fs_values[$field_name]);
          $this->customFieldsValues[$field_name] = reset($first_value);
        }
      }
    }
    parent::submitConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'transactionId' => NULL,
      'credit_card_number' => NULL,
      'expiration_date' => NULL,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getSupportedCurrencies() {
    return TRUE;
  }

  /**
   * Gets the ID of the payment method this plugin is for.
   *
   * @return string
   */
  public function getEntityId() {
    return $this->pluginDefinition['entity_id'];
  }

  /**
   * Gets the status to set on payment execution.
   *
   * @return string
   *   The plugin ID of the payment status to set.
   */
  public function getExecuteStatusId() {
    return $this->pluginDefinition['execute_status_id'];
  }

  /**
   * Gets the status to set on payment capture.
   *
   * @return string
   *   The plugin ID of the payment status to set.
   */
  public function getCaptureStatusId() {
    return $this->pluginDefinition['capture_status_id'];
  }

  /**
   * Gets whether or not capture is supported.
   *
   * @param bool
   *   Whether or not to support capture.
   */
  public function getCapture() {
    return $this->pluginDefinition['capture'];
  }

  /**
   * Gets the status to set on payment refund.
   *
   * @return string
   *   The plugin ID of the payment status to set.
   */
  public function getRefundStatusId() {
    return $this->pluginDefinition['refund_status_id'];
  }

  /**
   * Gets whether or not capture is supported.
   *
   * @param bool
   *   Whether or not to support capture.
   */
  public function getRefund() {
    return $this->pluginDefinition['refund'];
  }

  /**
   * Gets Authorize.net Environment endpoint.
   *
   * @return string
   */
  public function getApiEnvironment() {
    return $this->AuthnetProfile->getApiEnvironment();
  }

  protected function getAnetApiPayment() {
    return $this->AnetApiPayment;
  }

  protected function setAnetApiPayment() {
    if (!$this->AnetApiPayment) {
      $this->AnetApiPayment = new AnetAPI\PaymentType();
    }

    return $this;
  }

  /**
   * Sets Authorize.net Transaction Type.
   *
   * Sets $this->transactionType. Possible values:
   * - PAYMENT_AUTHNET_TRANSACTION_AUTHORIZE
   * - PAYMENT_AUTHNET_TRANSACTION_CAPTURE_AUTHORIZED
   * - PAYMENT_AUTHNET_TRANSACTION_CAPTURE
   * - PAYMENT_AUTHNET_TRANSACTION_REFUND
   *
   * @see payment_authnet.module
   * @see https://developer.authorize.net/api/reference/#payment-transactions
   * @see \AnetAPI\TransactionRequestType::setTransactionType().
   *
   * @param string $operation
   *  Payment Status ID to identify the operation that needs to occur.
   *  Possible values: payment_authorized, payment_success, payment_refunded.
   *
   * @return $this
   */
  protected function setTransactionType($operation) {
    try {
      $transaction_details = $this->getTransactionDetails();

      // Check if operation accepts money transfer.
      if (in_array($operation, ['payment_success', 'payment_money_transferred'])) {
        $this->transactionType = $transaction_details->getTransactionType() == PAYMENT_AUTHNET_TRANSACTION_AUTHORIZE
          ? PAYMENT_AUTHNET_TRANSACTION_CAPTURE_AUTHORIZED
          : PAYMENT_AUTHNET_TRANSACTION_CAPTURE;
      }

      // Check if status ancestor accepts money transfer.
      $ancestors = $this->getPayment()->getPaymentStatus()->getAncestors();
      if (in_array('payment_money_transferred', $ancestors)) {
        $this->transactionType = $transaction_details->getTransactionType() == PAYMENT_AUTHNET_TRANSACTION_AUTHORIZE
          ? PAYMENT_AUTHNET_TRANSACTION_CAPTURE_AUTHORIZED
          : PAYMENT_AUTHNET_TRANSACTION_CAPTURE;
      }

      // Check if this is refund operation.
      if ('payment_refunded' == $operation
        && in_array($transaction_details->getTransactionType(), [
          'authCaptureTransaction',
          'authOnlyTransaction',
        ])
        && in_array($transaction_details->getTransactionStatus(), [
          'settledSuccessfully',
        ])) {
        $this->transactionType = PAYMENT_AUTHNET_TRANSACTION_REFUND;
      }
    }
    catch (Exception $e) {
      \Drupal::logger('payment_authnet')->error($e->getMessasge());
    }

    return $this;
  }

  /**
   * Detects if authorize.net should transfer money right now.
   *
   * @return string
   *   Transaction type for Authorize net to identify will money be charged
   *   right now, or credit card will be just authorized.
   */
  public function getTransactionType() {
    return $this->transactionType;
  }

  /**
   * Creates the Authorize.net Transaction and sets needed props.
   *
   * @see \Drupal\payment_authnet\Plugin\Payment\Method\Authnet::getAnetApiTransaction()
   *
   * @return $this
   */
  protected function setAnetApiTransactionRequest() {
    if (!$this->transactionRequest) {
      $this->transactionRequest = new AnetAPI\TransactionRequestType();
    }
    $this->transactionRequest->setTransactionType($this->getTransactionType());
    $this->transactionRequest->setAmount($this->getPayment()->getAmount());
    $this->transactionRequest->setOrder($this->getAnetApiOrder());
    $this->transactionRequest->setLineItems($this->getAnetApiLineItems());
    $this->transactionRequest->setCustomer($this->getAnetApiCustomerData());
    $transaction_id = $this->getConfiguration('transactionId');
    // For authorized cards we don't need payment info but need Transaction ID.
    if ($transaction_id) {
      $this->transactionRequest->setRefTransId($transaction_id);
      // For refund operations we also need last 4 digits of credit card number
      // and expiration date.
      if ($this->getTransactionType() == PAYMENT_AUTHNET_TRANSACTION_REFUND) {
        $credit_card_number = $this->getConfiguration('credit_card_number');
        $expiration_date = $this->getConfiguration('expiration_date');
        $this->setCreditCard($credit_card_number, $expiration_date);
        $this->transactionRequest->setPayment($this->getAnetApiPayment());
      }
    }
    // For new payments set Payment data (authorize or authorize and capture
    // operations).
    else {
      $this->transactionRequest->setPayment($this->getAnetApiPayment());
    }

    return $this;
  }

  protected function setCreditCard($card_number = NULL, $expiration_date = NULL, $cvv = NULL) {
    $creditCard = new AnetAPI\CreditCardType();
    $creditCard->setCardNumber($card_number);
    $creditCard->setExpirationDate($expiration_date);
    $creditCard->setCardCode($cvv);

    $this->AnetApiPayment->setCreditCard($creditCard);
  }

  protected function getCreditCard() {
    return $this->getAnetApiPayment()->getCreditCard();
  }

  /**
   * Returns the list of line items for ANetApi Transaction.
   *
   * @return \net\authorize\api\contract\v1\LineItemType[]
   *  Array of objects.
   */
  protected function getAnetApiLineItems() {
    $line_items = [];
    foreach ($this->getPayment()->getLineItems() as $payment_line_item) {
      $line_item = new AnetAPI\LineItemType;
      $line_item->setItemId($payment_line_item->getPluginId());
      $line_item->setName(empty($payment_line_item->getName()) ? $payment_line_item->getPluginId() : $payment_line_item->getName());
      $line_item->setQuantity($payment_line_item->getQuantity());
      $line_item->setDescription($payment_line_item->getDescription());
      $line_item->setUnitPrice($payment_line_item->getAmount());
      $line_items[] = $line_item;
    }
    return $line_items;
  }

  /**
   * Gets the customer's identifying information.
   *
   * @return \net\authorize\api\contract\v1\CustomerDataType
   *   Customer Data for Anet Transaction.
   */
  protected function getAnetApiCustomerData() {
    $customerData = new AnetAPI\CustomerDataType();
    $customerData->setType(PAYMENT_AUTHNET_CUSTOMER_DATA_TYPE);
    $customerData->setId(strval($this->getPayment()->getOwner()->id()));
    $customerData->setEmail($this->getPayment()->getOwner()->getEmail());
    return $customerData;
  }

  /**
   * Creates order information.
   *
   * @see @see \Drupal\payment_authnet\Plugin\Payment\Method\Authnet::getAnetApiTransactionRequest()
   *
   * @return \net\authorize\api\contract\v1\OrderType
   */
  protected function getAnetApiOrder() {
    $order = new AnetAPI\OrderType();
    $order->setInvoiceNumber($this->getPayment()->id());

    return $order;
  }

  /**
   * Creates Authorize.net transaction.
   *
   * @see \Drupal\payment_authnet\Plugin\Payment\Method\Authnet::doExecutePayment()
   *
   * @return \net\authorize\api\contract\v1\TransactionRequestType
   */
  protected function getAnetApiTransactionRequest() {
    $this->setAnetApiTransactionRequest();

    // Add some merchant defined fields. These fields won't be stored with
    // the transaction, but will be echoed back in the response.
    // @todo :: Find the more accurate way to get field values.
    foreach ($this->customFieldsValues as $field_name => $field_value) {
      $merchantDefinedField = new AnetAPI\UserFieldType();
      $merchantDefinedField->setName($field_name);
      $merchantDefinedField->setValue(is_string($field_value) ? $field_value : Json::encode($field_value));
      $this->transactionRequest->addToUserFields($merchantDefinedField);
    }
    return $this->transactionRequest;
  }

  /**
   * Assemble the complete transaction request.
   *
   * @return object
   *   Transaction Response.
   *
   * @throws Exception\PaymentAuthnetApiException
   * @throws Exception\PaymentAuthnetResponseException
   */
  protected function sendRequest() {
    // Set the transaction's refId.
    $refId = $this->AuthnetProfile->getApiKey() . '_' . REQUEST_TIME;
    $request = new AnetAPI\CreateTransactionRequest();
    $request->setMerchantAuthentication($this->AuthnetProfile->getMerchantAuthentication());
    $request->setRefId($refId);
    $request->setTransactionRequest($this->getAnetApiTransactionRequest());
    // Create the controller and get the response.
    $controller = new AnetController\CreateTransactionController($request);
    $response = $controller->executeWithApiResponse($this->getApiEnvironment());

    if ($response->getRefId() != $refId) {
      throw new Exception\PaymentAuthnetResponseException($this->t('Invalid Reference ID in response. Error code: @code, Error message: @message', [
        '@code' => $response->getMessages()->getResultCode(),
        '@message' => $response->getMessages()->getMessage()[0]->getText(),
      ]));
    }
    $tresponse = $response->getTransactionResponse();
    // Check to see if the API request was not successful.
    if ($response->getMessages()->getResultCode() != 'Ok') {
      if ($tresponse->getErrors() != null) {
        throw new Exception\PaymentAuthnetResponseException($this->t('@message',[
          '@message' => $tresponse->getErrors()[0]->getErrorText(),
        ]), $tresponse->getErrors()[0]->getErrorCode());
      }
      else {
        throw new Exception\PaymentAuthnetResponseException($this->t('@status: @message.', [
          '@status' => $response->getMessages()->getResultCode(),
          '@message' => $response->getMessages()->getMessage()[0]->getText(),
        ]));
      }
    }

    return $tresponse;
  }

  /**
   *
   * @param string $operation
   *  Payment Status ID to identify the operation that needs to occur.
   *  Possible values: payment_authorized, payment_success, payment_refunded.
   *
   * @return boolean
   *   TRUE if operation was successful, FALSE otherwise.
   */
  protected function processResults($operation) {
    $this->setTransactionType($operation);
    try {
      $response = $this->sendRequest();
      // Since the API request was successful, look for a transaction response
      // and parse it to display the results of authorizing the card.
      $message_type = $response->getResponseCode() === '1' ? 'status' : 'warning';
      drupal_set_message($response->getMessages()[0]->getDescription(), $message_type);
      // Store the Authnet Transaction ID.
      if ($response->getTransId()) {
        $this->setConfiguration(['transactionId' => $response->getTransId()]);
      }
      return TRUE;
    }
    catch (Exception\PaymentAuthnetException $e) {
      drupal_set_message($e->getMessage(), 'error');
      \Drupal::logger('payment_authnet')->error($e->getMessage() . ' (' . $e->getCode() . ")|" . $e->getTraceAsString());
    }
    catch (Exception $e) {
      \Drupal::logger('payment_authnet')->error($e->getMessasge());
    }

    return FALSE;
  }

  protected function updateTransactionDetails() {

  }

  /**
   * Gets detailed information about a last Authorize.net transaction.
   *
   * @return \net\authorize\api\contract\v1\TransactionDetailsType
   *   Instance of getTransactionDetailsResponse.
   *
   * @throws Exception\PaymentAuthnetApiException
   * @throws Exception\PaymentAuthnetResponseException
   */
  public function getTransactionDetails() {
    if ($this->transactionDetails) {
      return $this->transactionDetails;
    }

    $this->transactionDetails = new AnetAPI\TransactionDetailsType();
    $transaction_id = $this->getConfiguration('transactionId');
    if (!empty($transaction_id)) {
      // Set the transaction's refId.
      $refId = $this->AuthnetProfile->getApiKey() . '_' . REQUEST_TIME;

      $request = new AnetAPI\GetTransactionDetailsRequest();
      $request->setMerchantAuthentication($this->AuthnetProfile->getMerchantAuthentication());
      $request->setRefId($refId);
      $request->setTransId($transaction_id);

      // Create the controller and get the response.
      $controller = new AnetController\GetTransactionDetailsController($request);
      $response = $controller->executeWithApiResponse($this->getApiEnvironment());
      if ($response->getRefId() != $refId) {
        throw new Exception\PaymentAuthnetResponseException($this->t('Invalid Reference ID in response. Error code: @code, Error message: @message', [
          '@code' => $response->getMessages()->getResultCode(),
          '@message' => $response->getMessages()->getMessage()[0]->getText(),
        ]));
      }
      if ($response->getMessages()->getResultCode() != 'Ok') {
        throw new Exception\PaymentAuthnetResponseException($this->t('@status: @message.', [
          '@status' => $response->getMessages()->getResultCode(),
          '@message' => $response->getMessages()->getMessage(),
        ]));
      }
      $this->transactionDetails = $response->getTransaction();
    }

    return $this->transactionDetails;
  }

  /**
   * {@inheritdoc}
   */
  protected function doExecutePayment() {
    if ($this->processResults($this->getExecuteStatusId())) {
      $this->getPayment()->setPaymentStatus($this->paymentStatusManager->createInstance($this->getExecuteStatusId()));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function doCapturePayment() {
    if ($this->processResults($this->getCaptureStatusId())) {
      $this->getPayment()->setPaymentStatus($this->paymentStatusManager->createInstance($this->getCaptureStatusId()));
      $this->getPayment()->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function doCapturePaymentAccess(AccountInterface $account) {
    return $this->getCapture()
    && $this->getPayment()->getPaymentStatus()->getPluginId() != $this->getCaptureStatusId();
  }

  /**
   * {@inheritdoc}
   */
  public function doRefundPayment() {
    if ($this->processResults($this->getRefundStatusId())) {
      $this->getPayment()->setPaymentStatus($this->paymentStatusManager->createInstance($this->getRefundStatusId()));
      $this->getPayment()->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function doRefundPaymentAccess(AccountInterface $account) {
    return
      $this->getRefund()
      && $this->getPayment()->getPaymentStatus()->getPluginId() != $this->getRefundStatusId()
      && $this->getConfiguration('credit_card_number')
      && $this->getConfiguration('expiration_date')
      && $this->getConfiguration('transactionId')
      && $this->getTransactionDetails()->getTransactionStatus() == 'settledSuccessfully';
  }

  /**
   * {@inheritdoc}
   */
  public function updatePaymentStatusAccess(AccountInterface $account) {
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function getSettablePaymentStatuses(AccountInterface $account, PaymentInterface $payment) {
    return [
      'payment_pending',
      'payment_success',
      'payment_authorized',
      'payment_refunded',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    $this->configuration = $configuration + $this->configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function getConfiguration($key = NULL) {
    if ($key) {
      return empty($this->configuration[$key]) ? NULL : $this->configuration[$key];
    }
    return $this->configuration;
  }
}
